""" Webcropper for uni-potsdam.de/nachrichten """
# TESTED 15th of November 2021
# !!RUNTIME 4h 19min
# pylint: disable=duplicate-code

import os
import logging
import time
import requests as rq
from bs4 import BeautifulSoup as bsx

from cropping.models import Category, Article, Meta
from one_api.settings_prod import APP_DATA_ROOT
from articles.models import Medium, MediumType

logger = logging.getLogger(__name__)


class MetaFile:
    """ stores the data from metafile """
    is_joining = False
    cath_index = 0  # category index by array position
    page_index = 1  # page number (starts at 1)
    art_index = 0  # article index (restarts count every page)
    wait_index = 0

    def save(self):
        """ updating the pseudometa file according to the objects data """
        with open(APP_DATA_ROOT + "pseudometa.txt", "w+", encoding='UTF-8') as psdmeta:
            psdmeta.write(str(self.cath_index) + ',' + str(self.page_index) + ',' + str(self.art_index) + ',' + str(
                int(self.is_joining)))


# helper
def get_scoop(rq_url, rq_local):
    """ request and souping for either local or web pages """
    if rq_local:
        res = ""
        with open(rq_url, 'r', encoding='UTF-8') as page:
            res += page.read()
        return bsx(res, 'html.parser')
    res = rq.get(rq_url)
    return bsx(res.text, 'html.parser')


def reset_join(mode):
    """ reset the pseudometa either to joining mode (mode=True) or to restart (mode=False) """
    with open(APP_DATA_ROOT + "pseudometa.txt", "w+", encoding='UTF-8') as psdmeta:
        psdmeta.write("0,1,0," + str(int(mode)))


def get_categories(branch_prefix, branch_sets):
    """ getting categories """
    # setup
    logger.info("starting the cropper")
    ls_curls = []

    # squeeze available categories
    logger.debug("squeeze available categories:")
    scoop = get_scoop(branch_prefix + str(1) + branch_sets + str(1), False)
    c_urls = list(map(lambda x: x['href'], scoop.findAll('a')))  # getting the category URLs
    c_urls = list(filter(lambda x: x.startswith("/de/nachrichten/page?tx_news_pi"), c_urls))
    c_names = list(filter(lambda x: x.text.startswith(" "), scoop.findAll('a')))  # getting names
    for url, name in zip(c_urls, c_names):
        ls_curls.append((url[70:], name.text[1: len(name.text) - 1]))
        cat_name = str(name.text[1: len(name.text) - 1])
        cat, created = Category.objects.update_or_create(name=cat_name)
        if created:
            logger.debug('saved category %s (%s)', cat.name, cat.pk)

    logger.debug(str(len(ls_curls)) + " categories found")
    return ls_curls


def run_crop(categories, branch, art_local, crop_session, meta):
    """ processing categories """
    for dv_url in categories[meta.cath_index:]:
        logger.debug("gathering on topic: " + dv_url[1])
        ls_urls, branch_page = [], meta.page_index

        # getting relevant URLs
        extract_relevant_urls(art_local, branch, branch_page, dv_url, ls_urls)

        # work with found articles
        logger.debug("starting to read the found articles")
        read_articles(ls_urls, dv_url, art_local, crop_session, meta)

        # prepare meta for next category
        meta.cath_index += 1
        meta.page_index = 1
        meta.art_index = 0


def extract_relevant_urls(art_local, branch, branch_page, dv_url, ls_urls):
    """ extracting relevant urls from a html site """
    while True:
        logger.debug("getting available links on page " + str(branch_page))
        logger.info(dv_url[0])
        branch_url = branch[0] + str(branch_page) + branch[1] + dv_url[0]
        if art_local:
            scoop = get_scoop(dv_url[0], True)
            t_urls = map(str, scoop.findAll('a'))
            t_urls = list(filter(lambda x: x.startswith("<a href"), t_urls))
            t_urls = list(filter(lambda x: "html" in x, t_urls))
            t_urls = list(map(lambda x: x.split('href="', 1)[1], t_urls))
            t_urls = list(map(lambda x: x.split('"', 1)[0], t_urls))
            t_urls = list(filter(lambda x: x.startswith("/"), t_urls))
            t_urls = list(dict.fromkeys(t_urls))
            url_prefix = os.getcwd()
        else:
            scoop = get_scoop(branch_url, False)
            t_urls = list(map(lambda x: x['href'], scoop.findAll('a')))
            t_urls = list(filter(lambda x: x.startswith("/de/nachrichten/detail"), t_urls))
            url_prefix = 'https://www.uni-potsdam.de'
        if not t_urls:
            break
        t_last = ""
        for t_url in t_urls:
            if t_last != t_url:
                ls_urls.append(url_prefix + t_url)
            t_last = t_url
        branch_page += 1
        if art_local:
            break


def read_articles(articles, category, art_local, crop_session, meta):
    """ read and process articles """
    for url in get_url(articles, meta):
        logger.info("now reading: " + url)

        s_article = crop_html(art_local, category, crop_session, meta, url)

        # write to csv
        write_categories_to_file(category, s_article)

        # update meta file
        meta.art_index += 1
        if meta.art_index > 9:
            meta.page_index += 1
            meta.art_index = 0
        meta.save()


def write_categories_to_file(category, s_article):
    """ writes the given categories into export_raw.csv """
    with open(APP_DATA_ROOT + "export_raw.csv", "a", encoding='UTF-8') as tch:
        tch.write(category[1] + "," + s_article + "\n")


def get_url(articles, meta):
    """ get url from list """
    return articles[meta.art_index:]


def crop_html(art_local, category, crop_session, meta, url):
    """ extract title and relevant text from html """
    # respectfully waiting if necessary
    if meta.wait_index > 200:
        logger.debug("respectfully waiting for 1 minutes...")
        time.sleep(60 * int(not art_local))
        meta.wait_index = 0
    else:
        meta.wait_index += 1
    # compiling article
    scoop, s_article = get_scoop(url, art_local), ""
    up_content = scoop.find(id='up_content')
    title = up_content.find('h1')
    paragraphs = up_content.findAll('p')
    for paragraph in paragraphs:
        s_article += " ".join(paragraph.text.split()) + "\n"
    title_text = " ".join(title.text.strip().split())
    article, _ = Article.objects.update_or_create(title=title_text, session_id=crop_session.id,
                                                  defaults={'text': s_article, 'medium': get_or_create_medium()})
    logger.debug("article saved: %i", article.id)
    s_article = title_text + "\n\n" + s_article
    cat_obj, _ = Category.objects.update_or_create(name=str(category[1]))
    logger.debug("added category %s to article %i", cat_obj.name, article.id)
    article.categories.add(cat_obj)
    article.save()
    return s_article


def get_or_create_medium():
    """ Takes the first available medium or creates a new one if none exist """
    medium = Medium.objects.all()[0]
    if isinstance(medium, Medium):
        return medium

    medium_type = MediumType(name='Cropper Default Type', description='Cropper Default Type')
    medium_type.save()
    default_medium = Medium.objects.create(name='Cropper Default', medium_type=medium_type)
    return default_medium


def join_categories():
    """ join duplicated articles from different categories """
    logger.info("searching & joining duplicates...")
    with open(APP_DATA_ROOT + "export_raw.csv", "r", encoding='UTF-8') as rfile:
        lines = list(map(lambda x: x.split(',', 1), rfile.readlines()))
        join_categories_in_lines(lines)

    # rewrite joined information as updated csv
    logger.info("writing the final .csv file...")
    with open("export.csv", "a", encoding='UTF-8') as rwr:
        for line in lines:
            rwr.write(line[0] + ',' + line[1])


def join_categories_in_lines(lines):
    """ join duplicated categories in line """
    for j, line in enumerate(lines):
        i = j + 1
        while i < len(lines):
            if line[1] == lines[i][1]:
                logger.debug("joining duplicate between \"" + line[0] + "\" and \"" + lines[i][0] + "\"")
                line[0] += ';' + lines[i][0]
                lines.pop(i)
            else:
                i += 1


def start_crop(prefix, sets, crop_session, restart):
    """ runs the entire cropping process """

    # load meta file to finish where last left off
    if Meta.objects.count() > 0:
        meta = Meta.objects.all()[0]
    else:
        meta = Meta()
        meta.save()
    if not restart:
        reset_join(False)  # if file wasn't created yet

    # run file cropping if not jet in joining phase
    if not meta.is_joining:
        run_crop(get_categories(prefix, sets), (prefix, sets), False, crop_session, meta)
        reset_join(True)

    join_categories()
    reset_join(False)

# TODO: automatically create a pseudometa if not available in desired path
