""" Trains the model, that is later used for prediction of the articles """

import logging
import nltk
from nltk.corpus import stopwords
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.multiclass import OneVsRestClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.pipeline import make_pipeline
from sklearn.pipeline import Pipeline
import pandas as pd
from django.utils import timezone
# from prediction.models import PredSession, TrainedModels
from cropping.models import Article

logger = logging.getLogger(__name__)


# def predict_article(articles_pred, articles_train):
def predict_article(_, articles_train):
    """Prediction of article labels"""

    _, labels, _, train, _, x_train, _, _, nb_pipeline = define_pipeline(articles_train)
    logger.info('Prediction of articles')

    # predicted_labels = []

    # p = Paginator(articles, 100)
    # pred_session = PredSession(date_of_pred=timezone.now())

    # if p.count != 0:
    for label in labels:
        # for article in articles_pred:
        logger.info('... Processing %s', label)
        # train the model using X_dtm & y
        nb_pipeline.fit(x_train, train[str(label)])
        logger.info('Pipeline gefittet')

    id_label_list = []

    for article in Article.objects.filter(id__lte=30):
        for label in labels:
            logger.info('... Processing %s', label)
            single_article_df = pd.DataFrame(data=[article.text])
            predicted_label = nb_pipeline.predict(single_article_df.iloc[0])
            logger.info('Predicted Label: ' + str(predicted_label))
            logger.info(timezone.now())
            if predicted_label == 1:
                logger.info('Artikel ' + str(article.id) + ' hat das Label ' + str(label))
                id_label_list.append([article.id, str(label)])
                # TrainedModels.objects.update_or_create(articles_id=article.id, pred_label=str(label), train_session=pred_session)
                logger.info('Saved to database')

    logger.info(id_label_list)
    logger.info('Prediction done.')
    # return predicted_labels


def define_pipeline(articles):
    """read data and define pipeline"""
    # articles are dataframe of [text, categories]

    # set axis
    if len(articles.columns) > 1:
        articles.set_axis(['Labels', 'Article_text'], axis=1, inplace=True)

    # get labels list
    labels = get_label_list(articles)

    # create list of 0 and 1
    splitted_list = create_0_1_list(articles)

    # append the data by the 0 1 list
    data = append_dataframe_by_0_1_list_entries(articles)

    # predict
    # create train and test set
    train, test = train_test_split(data, random_state=42, test_size=0.2, shuffle=True)
    x_train = train.Article_text
    x_test = test.Article_text

    # Define a pipeline with multi lable classifier
    nltk.download('stopwords')
    stop_words = set(stopwords.words('german'))

    nb_pipeline = Pipeline([
        ('tfidf', TfidfVectorizer(stop_words=stop_words)),
        ('clf', OneVsRestClassifier(MultinomialNB(
            fit_prior=True, class_prior=None))),
    ])

    logger.info('Pipeline gebildet')

    return data, labels, splitted_list, train, test, x_train, x_test, stop_words, nb_pipeline


def train_model_multiple_labels(articles):
    """ Train the model and predict the labels"""
    _, labels, _, train, test, x_train, x_test, _, nb_pipeline = define_pipeline(articles)

    # fit the data and predict for each label
    # article has to be a dataframe
    predicted_list = []
    for label in labels:
        logger.info('... Processing %s', label)
        # train the model
        nb_pipeline.fit(x_train, train[str(label)])
        # predict the labels and compute the accuracy
        # will predict 0 or 1 for the articles
        prediction = nb_pipeline.predict(x_test)
        predicted_list.append(prediction)
        logger.info(prediction)
        logger.info('Test accuracy is %s', accuracy_score(test[str(label)], prediction))

    logger.info('Prediction done.')

    output = 'Training Multiple done successfully'
    logger.info(output)


def train_model_single_labels(file_path):
    """ Train a model for single labels """
    # load the data with single label
    data_single = pd.read_csv(file_path, header=None,
                              sep='^([^,]+),', engine='python')
    data_single = data_single.iloc[:, 1:3]
    article_text_2 = data_single.iloc[:, 1]
    article_labels_2 = data_single.iloc[:, 0]

    # predict
    # train the model with Multinomial Naive Bayes

    # define model
    model = make_pipeline(TfidfVectorizer(), MultinomialNB())

    # split data into train and test set
    x_train, _, y_train, _ = train_test_split(article_text_2, article_labels_2, test_size=0.3)
    # , random_state=42)

    # fit the data to the model
    model.fit(x_train, y_train)

    # predict labels
    # pred_labels = model.predict(x_test[0:10])
    # output = 'Training Single Label done successfully'


# only unsed when we have a csv not the data in the database
def read_data(file_path):
    """ Read the data and set axis """
    # load the data with first column: labels and second column: article text
    data = pd.read_csv(file_path, header=None,
                       sep='^([^,]+),', engine='python')
    data = data.iloc[:, 1:3]
    print(len(data))
    data.set_axis(['Article_text', 'Labels'], axis=1, inplace=True)

    return data


def get_label_list(data):
    """ get all labels that are assigned to the articles """
    # define labels
    labels = []
    label_data_list = []

    # prepare labels
    label_data = list(data.iloc[:, 0])

    # convert the label_data entries to arrays of strings
    for element in enumerate(label_data):
        label_data_list.append(element)

    # loop over the array and append labels without duplicates
    for labeled_data_list_element in enumerate(label_data_list):
        for element in enumerate(labeled_data_list_element):
            if element not in labels:
                labels.append(element)

    logger.info('Das sind die Label: ' + str(labels))

    logger.info(labels)
    return labels


def create_0_1_list(data):
    """ create a list that has 0 if the article does not have the label and 1 if it has the label """
    # get labels
    labels = get_label_list(data)

    # create list to save 0 and 1
    list_0_1 = []

    # check if article is assigned to label
    for i in range(len(data)):
        for label in labels:
            if label in data.iloc[i, 0]:
                list_0_1.append(1)
            else:
                list_0_1.append(0)

    # function to split list into parts
    def split_list(alist):
        splitted_list_all_entries = []
        count_labels = len(labels)  # 8 label
        for k in range(len(data)):
            splitted_list_all_entries.append(alist[(k * count_labels): ((k + 1) * count_labels)])

        return splitted_list_all_entries

    # seperate the list into smaller lists with 110 entries each
    splitted_list = split_list(list_0_1)
    return splitted_list


def append_dataframe_by_0_1_list_entries(data):
    """ use the entries to append the dataframe """
    # get labels
    labels = get_label_list(data)

    # get splitted list
    splitted_list = create_0_1_list(data)

    # create empty lists
    lists = []
    for i, _ in enumerate(labels):
        lists.append([])

    # create lists that have the data for each label
    for i, _ in enumerate(splitted_list):
        for j, _ in enumerate(labels):
            lists[j].append(splitted_list[i][j])

    # append dataframe by the lists
    for k, element in enumerate(labels):
        data.insert(k + 2, str(element), lists[k])

    return data
